from __future__ import unicode_literals
# flake8: noqa
import os
import sys

from collections import OrderedDict, defaultdict
from fnmatch import fnmatch
from importlib import import_module

import django
import six
from django.core.management.base import BaseCommand, CommandError
import django.template
from django.template import Context
from django.utils.encoding import smart_text
from django.template.loader import get_template  # noqa Leave this in to preload template locations
from django.template import engines

from compressor.cache import get_offline_hexdigest, write_offline_manifest, get_offline_manifest
from compressor.conf import settings
from compressor.exceptions import (OfflineGenerationError, TemplateSyntaxError,
                                   TemplateDoesNotExist)
from compressor.utils import get_mod_func


class Command(BaseCommand):
    help = "Compress content outside of the request/response cycle"

    def add_arguments(self, parser):
        parser.add_argument('--extension', '-e', action='append', dest='extensions',
                            help='The file extension(s) to examine (default: ".html", '
                                 'separate multiple extensions with commas, or use -e '
                                 'multiple times)')
        parser.add_argument('-f', '--force', default=False, action='store_true',
                            help="Force the generation of compressed content even if the "
                                 "COMPRESS_ENABLED setting is not True.", dest='force')
        parser.add_argument('--follow-links', default=False, action='store_true',
                            help="Follow symlinks when traversing the COMPRESS_ROOT "
                                 "(which defaults to STATIC_ROOT). Be aware that using this "
                                 "can lead to infinite recursion if a link points to a parent "
                                 "directory of itself.", dest='follow_links')
        parser.add_argument('--engine', default=[], action="append",
                            help="Specifies the templating engine. jinja2 and django are "
                                 "supported. It may be a specified more than once for "
                                 "multiple engines. If not specified, django engine is used.",
                            dest="engines")

    def get_loaders(self):
        template_source_loaders = []
        for e in engines.all():
            if hasattr(e, 'engine'):
                template_source_loaders.extend(
                    e.engine.get_template_loaders(e.engine.loaders))
        loaders = []
        # If template loader is CachedTemplateLoader, return the loaders
        # that it wraps around. So if we have
        # TEMPLATE_LOADERS = (
        #    ('django.template.loaders.cached.Loader', (
        #        'django.template.loaders.filesystem.Loader',
        #        'django.template.loaders.app_directories.Loader',
        #    )),
        # )
        # The loaders will return django.template.loaders.filesystem.Loader
        # and django.template.loaders.app_directories.Loader
        # The cached Loader and similar ones include a 'loaders' attribute
        # so we look for that.
        for loader in template_source_loaders:
            if hasattr(loader, 'loaders'):
                loaders.extend(loader.loaders)
            else:
                loaders.append(loader)
        return loaders

    def __get_parser(self, engine):
        charset = (
            settings.FILE_CHARSET if settings.is_overridden('FILE_CHARSET')
            else 'utf-8'
        )
        if engine == "jinja2":
            from compressor.offline.jinja2 import Jinja2Parser
            env = settings.COMPRESS_JINJA2_GET_ENVIRONMENT()
            parser = Jinja2Parser(charset=charset, env=env)
        elif engine == "django":
            from compressor.offline.django import DjangoParser
            parser = DjangoParser(charset=charset)
        else:
            raise OfflineGenerationError("Invalid templating engine specified.")

        return parser

    def compress(self, engine, extensions, verbosity, follow_links, log):
        """
        Searches templates containing 'compress' nodes and compresses them
        "offline" -- outside of the request/response cycle.

        The result is cached with a cache-key derived from the content of the
        compress nodes (not the content of the possibly linked files!).
        """

        if not self.get_loaders():
            raise OfflineGenerationError("No template loaders defined. You "
                                         "must set TEMPLATE_LOADERS in your "
                                         "settings or set 'loaders' in your "
                                         "TEMPLATES dictionary.")
        templates = set()
        if engine == 'django':
            paths = set()
            for loader in self.get_loaders():
                try:
                    module = import_module(loader.__module__)
                    get_template_sources = getattr(module,
                        'get_template_sources', None)
                    if get_template_sources is None:
                        get_template_sources = loader.get_template_sources
                    paths.update(smart_text(origin) for origin in get_template_sources(''))
                except (ImportError, AttributeError, TypeError):
                    # Yeah, this didn't work out so well, let's move on
                    pass

            if not paths:
                raise OfflineGenerationError("No template paths found. None of "
                                             "the configured template loaders "
                                             "provided template paths. See "
                                             "https://docs.djangoproject.com/en/2.1/topics/templates/ "
                                             "for more information on template "
                                             "loaders.")
            if verbosity >= 2:
                log.write("Considering paths:\n\t" + "\n\t".join(paths) + "\n")

            for path in paths:
                for root, dirs, files in os.walk(path, followlinks=follow_links):
                    templates.update(os.path.join(root, name)
                        for name in files if not name.startswith('.') and
                            any(fnmatch(name, "*%s" % glob) for glob in extensions))
        elif engine == 'jinja2':
            env = settings.COMPRESS_JINJA2_GET_ENVIRONMENT()
            if env and hasattr(env, 'list_templates'):
                templates |= set([env.loader.get_source(env, template)[1] for template in
                            env.list_templates(filter_func=lambda _path:
                            os.path.splitext(_path)[-1] in extensions)])

        if not templates:
            raise OfflineGenerationError("No templates found. Make sure your "
                                         "TEMPLATE_LOADERS and TEMPLATE_DIRS "
                                         "settings are correct.")
        if verbosity >= 2:
            log.write("Found templates:\n\t" + "\n\t".join(templates) + "\n")

        contexts = settings.COMPRESS_OFFLINE_CONTEXT
        if isinstance(contexts, six.string_types):
            try:
                module, function = get_mod_func(contexts)
                contexts = getattr(import_module(module), function)()
            except (AttributeError, ImportError, TypeError) as e:
                raise ImportError("Couldn't import offline context function %s: %s" %
                                  (settings.COMPRESS_OFFLINE_CONTEXT, e))
        elif not isinstance(contexts, (list, tuple)):
            contexts = [contexts]

        parser = self.__get_parser(engine)
        fine_templates = []

        if verbosity >= 1:
            log.write("Compressing... ")

        for template_name in templates:
            try:
                template = parser.parse(template_name)
                template.template_name = template_name
                fine_templates.append(template)
            except IOError:  # unreadable file -> ignore
                if verbosity >= 1:
                    log.write("Unreadable template at: %s\n" % template_name)
                continue
            except TemplateSyntaxError as e:  # broken template -> ignore
                if verbosity >= 1:
                    log.write("Invalid template %s: %s\n" % (template_name, smart_text(e)))
                continue
            except TemplateDoesNotExist:  # non existent template -> ignore
                if verbosity >= 1:
                    log.write("Non-existent template at: %s\n" % template_name)
                continue
            except UnicodeDecodeError:
                if verbosity >= 1:
                    log.write("UnicodeDecodeError while trying to read "
                              "template %s\n" % template_name)
                continue

        contexts_count = 0
        nodes_count = 0
        block_count = 0
        offline_manifest = OrderedDict()
        results = []
        for context_dict in contexts:
            compressor_nodes = OrderedDict()
            for template in fine_templates:
                context = Context(parser.get_init_context(context_dict))

                try:
                    nodes = list(parser.walk_nodes(template, context=context))
                except (TemplateDoesNotExist, TemplateSyntaxError) as e:
                    # Could be an error in some base template
                    if verbosity >= 1:
                        log.write("Error parsing template %s: %s\n" %
                                  (template.template_name, smart_text(e)))
                    continue

                if nodes:
                    template_nodes = compressor_nodes.setdefault(template, OrderedDict())
                    for node in nodes:
                        nodes_count += 1
                        template_nodes.setdefault(node, []).append(context)

            for template, nodes in compressor_nodes.items():
                template._log = log
                template._log_verbosity = verbosity

                for node, node_contexts in nodes.items():
                    for context in node_contexts:
                        context.push()
                        if not parser.process_template(template, context):
                            continue

                        parser.process_node(template, context, node)
                        rendered = parser.render_nodelist(template, context, node)
                        key = get_offline_hexdigest(rendered)

                        if key in offline_manifest:
                            continue

                        try:
                            result = parser.render_node(template, context, node)
                        except Exception as e:
                            raise CommandError("An error occurred during rendering %s: "
                                               "%s" % (template.template_name, smart_text(e)))
                        result = result.replace(
                            settings.COMPRESS_URL, settings.COMPRESS_URL_PLACEHOLDER
                        )
                        offline_manifest[key] = result
                        context.pop()
                        results.append(result)
                        block_count += 1

        if not nodes_count:
            raise OfflineGenerationError(
                "No 'compress' template tags found in templates."
                "Try running compress command with --follow-links and/or"
                "--extension=EXTENSIONS")

        if verbosity >= 1:
            log.write("done\nCompressed %d block(s) from %d template(s) for %d context(s).\n" %
                      (block_count, nodes_count, contexts_count))
        return offline_manifest, block_count, results

    def handle_extensions(self, extensions=('html',)):
        """
        organizes multiple extensions that are separated with commas or
        passed by using --extension/-e multiple times.

        for example: running 'django-admin compress -e js,txt -e xhtml -a'
        would result in an extension list: ['.js', '.txt', '.xhtml']

        >>> handle_extensions(['.html', 'html,js,py,py,py,.py', 'py,.py'])
        ['.html', '.js']
        >>> handle_extensions(['.html, txt,.tpl'])
        ['.html', '.tpl', '.txt']
        """
        ext_list = []
        for ext in extensions:
            ext_list.extend(ext.replace(' ', '').split(','))
        for i, ext in enumerate(ext_list):
            if not ext.startswith('.'):
                ext_list[i] = '.%s' % ext_list[i]
        return set(ext_list)

    def handle(self, **options):
        self.handle_inner(**options)

    def handle_inner(self, **options):
        if not settings.COMPRESS_ENABLED and not options.get("force"):
            raise CommandError(
                "Compressor is disabled. Set the COMPRESS_ENABLED "
                "setting or use --force to override.")
        if not settings.COMPRESS_OFFLINE:
            if not options.get("force"):
                raise CommandError(
                    "Offline compression is disabled. Set "
                    "COMPRESS_OFFLINE or use the --force to override.")

        log = options.get("log", sys.stdout)
        verbosity = options.get("verbosity", 1)
        follow_links = options.get("follow_links", False)
        extensions = self.handle_extensions(options.get("extensions") or ["html"])
        engines = [e.strip() for e in options.get("engines", [])] or ["django"]

        final_offline_manifest = {}
        final_block_count = 0
        final_results = []
        for engine in engines:
            offline_manifest, block_count, results = self.compress(engine, extensions, verbosity, follow_links, log)
            final_results.extend(results)
            final_block_count += block_count
            final_offline_manifest.update(offline_manifest)
        write_offline_manifest(final_offline_manifest)
        return final_block_count, final_results

Command.requires_system_checks = False



from __future__ import absolute_import, unicode_literals
import six

from django.core.exceptions import ImproperlyConfigured

from compressor.filters import FilterBase, CallbackOutputFilter


class rJSMinFilter(CallbackOutputFilter):
    callback = "rjsmin.jsmin"
    dependencies = ["rjsmin"]
    kwargs = {
        "keep_bang_comments": True
    }


# This is for backwards compatibility
JSMinFilter = rJSMinFilter


class SlimItFilter(CallbackOutputFilter):
    dependencies = ["slimit"]
    callback = "slimit.minify"
    kwargs = {
        "mangle": True,
    }


class CalmjsFilter(FilterBase):
    def __init__(self, *args, **kwargs):
        try:
            self._parser = kwargs.pop('parser')
        except KeyError:
            self._parser = None
        try:
            self._unparser = kwargs.pop('unparser')
        except KeyError:
            self._unparser = None
        super(CalmjsFilter, self).__init__(*args, **kwargs)
        try:
            import calmjs.parse
        except ImportError:
            raise ImproperlyConfigured(
                "The module calmjs.parse couldn't be imported. "
                "Make sure it is correctly installed.")
        if self._parser is None:
            self._parser = calmjs.parse.es5
        if self._unparser is None:
            self._unparser = calmjs.parse.unparsers.es5.minify_printer(obfuscate=True)

    def output(self, **kwargs):
        program = self._parser(self.content)
        minified = u''.join(part.text for part in self._unparser(program))
        assert isinstance(minified, six.text_type)
        return minified



from __future__ import with_statement, unicode_literals
import errno
import os
import brotli

from django.core.files.base import ContentFile
from django.core.files.storage import get_storage_class
from django.test import TestCase
from django.test.utils import override_settings
from django.utils.functional import LazyObject

from compressor import storage
from compressor.conf import settings
from compressor.tests.test_base import css_tag
from compressor.tests.test_templatetags import render


class GzipStorage(LazyObject):
    def _setup(self):
        self._wrapped = get_storage_class('compressor.storage.GzipCompressorFileStorage')()


class BrotliStorage(LazyObject):
    def _setup(self):
        self._wrapped = get_storage_class('compressor.storage.BrotliCompressorFileStorage')()


@override_settings(COMPRESS_ENABLED=True)
class StorageTestCase(TestCase):
    def setUp(self):
        self.default_storage = storage.default_storage
        storage.default_storage = GzipStorage()
        storage.brotli_storage = BrotliStorage()

    def tearDown(self):
        storage.default_storage = self.default_storage

    def test_gzip_storage(self):
        storage.default_storage.save('test.txt', ContentFile('yeah yeah'))
        self.assertTrue(os.path.exists(os.path.join(settings.COMPRESS_ROOT, 'test.txt')))
        self.assertTrue(os.path.exists(os.path.join(settings.COMPRESS_ROOT, 'test.txt.gz')))

    def test_brotli_storage(self):
        payload = ','.join([str(i) for i in range(1000)]).encode()
        chunk_size = 1024
        storage.brotli_storage.save('test.txt', ContentFile(payload))
        self.assertTrue(os.path.exists(os.path.join(settings.COMPRESS_ROOT, 'test.txt')))
        self.assertTrue(os.path.exists(os.path.join(settings.COMPRESS_ROOT, 'test.txt.br')))
        decompressed_data = b''
        br_decompressor = brotli.Decompressor()
        with open(os.path.join(settings.COMPRESS_ROOT, 'test.txt.br'), 'rb') as f:
            for data in iter(lambda: f.read(chunk_size), b''):
                decompressed_data += br_decompressor.process(data)
        self.assertEqual(payload, decompressed_data)

    def test_css_tag_with_storage(self):
        template = """{% load compress %}{% compress css %}
        <link rel="stylesheet" href="{{ STATIC_URL }}css/one.css" type="text/css">
        <style type="text/css">p { border:5px solid white;}</style>
        <link rel="stylesheet" href="{{ STATIC_URL }}css/two.css" type="text/css">
        {% endcompress %}
        """
        context = {'STATIC_URL': settings.COMPRESS_URL}
        out = css_tag("/static/CACHE/css/output.aca9bcd16bee.css")
        self.assertEqual(out, render(template, context))

    def test_race_condition_handling(self):
        # Hold on to original os.remove
        original_remove = os.remove

        def race_remove(path):
            "Patched os.remove to raise ENOENT (No such file or directory)"
            original_remove(path)
            raise OSError(errno.ENOENT, 'Fake ENOENT')

        try:
            os.remove = race_remove
            self.default_storage.save('race.file', ContentFile('Fake ENOENT'))
            self.default_storage.delete('race.file')
            self.assertFalse(self.default_storage.exists('race.file'))
        finally:
            # Restore os.remove
            os.remove = original_remove



from django.core.management import call_command
from django.core.management.base import CommandError
from django.test import TestCase
from six import StringIO


class TestMtimeCacheCommand(TestCase):
    # FIXME: add actual tests, improve the existing ones.

    exclusion_patterns = [
        '*CACHE*', '*custom*', '*066cd253eada.js', '*d728fc7f9301.js', '*8a0fed36c317.js', 'test.txt*'
    ]

    def default_ignore(self):
        return ['--ignore=%s' % pattern for pattern in self.exclusion_patterns]

    def test_handle_no_args(self):
        with self.assertRaises(CommandError):
            call_command('mtime_cache')

    def test_handle_add(self):
        out = StringIO()
        with self.settings(CACHES={}):
            call_command(
                'mtime_cache', '--add', *self.default_ignore(), stdout=out)
        output = out.getvalue()
        self.assertIn('Deleted mtimes of 20 files from the cache.', output)
        self.assertIn('Added mtimes of 20 files to cache.', output)

    def test_handle_clean(self):
        out = StringIO()
        with self.settings(CACHES={}):
            call_command(
                'mtime_cache', '--clean', *self.default_ignore(), stdout=out)
        output = out.getvalue()
        self.assertIn('Deleted mtimes of 20 files from the cache.', output)
        self.assertNotIn('Added mtimes of 20 files to cache.', output)



from django.template import Template, Context
from django.conf import settings

from compressor.filters import FilterBase


class TemplateFilter(FilterBase):

    def input(self, filename=None, basename=None, **kwargs):
        template = Template(self.content)
        context = Context(settings.COMPRESS_TEMPLATE_FILTER_CONTEXT)
        return template.render(context)



from compressor.conf import settings
from compressor.filters import CompilerFilter


class YUICompressorFilter(CompilerFilter):
    command = "{binary} {args}"

    def __init__(self, *args, **kwargs):
        super(YUICompressorFilter, self).__init__(*args, **kwargs)
        self.command += ' --type=%s' % self.type
        if self.verbose:
            self.command += ' --verbose'


class YUICSSFilter(YUICompressorFilter):
    type = 'css'
    options = (
        ("binary", settings.COMPRESS_YUI_BINARY),
        ("args", settings.COMPRESS_YUI_CSS_ARGUMENTS),
    )


class YUIJSFilter(YUICompressorFilter):
    type = 'js'
    options = (
        ("binary", settings.COMPRESS_YUI_BINARY),
        ("args", settings.COMPRESS_YUI_JS_ARGUMENTS),
    )



from __future__ import unicode_literals
import os
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.template.utils import InvalidTemplateEngineError

from appconf import AppConf


default_filters = dict(
    css=['compressor.filters.css_default.CssAbsoluteFilter'],
    js=['compressor.filters.jsmin.JSMinFilter'])


class CompressorConf(AppConf):
    # Main switch
    ENABLED = not settings.DEBUG
    # Allows changing verbosity from the settings.
    VERBOSE = False
    # GET variable that disables compressor e.g. "nocompress"
    DEBUG_TOGGLE = None
    # the backend to use when parsing the JavaScript or Stylesheet files
    PARSER = 'compressor.parser.AutoSelectParser'
    OUTPUT_DIR = 'CACHE'
    STORAGE = 'compressor.storage.CompressorFileStorage'

    COMPRESSORS = dict(
        css='compressor.css.CssCompressor',
        js='compressor.js.JsCompressor',
    )

    URL = None
    ROOT = None

    # Filters are resolved in configure()
    FILTERS = {}
    CSS_FILTERS = None
    JS_FILTERS = None

    CSS_HASHING_METHOD = 'mtime'

    PRECOMPILERS = (
        # ('text/coffeescript', 'coffee --compile --stdio'),
        # ('text/less', 'lessc {infile} {outfile}'),
        # ('text/x-sass', 'sass {infile} {outfile}'),
        # ('text/stylus', 'stylus < {infile} > {outfile}'),
        # ('text/x-scss', 'sass --scss {infile} {outfile}'),
    )
    CACHEABLE_PRECOMPILERS = ()
    CLOSURE_COMPILER_BINARY = 'java -jar compiler.jar'
    CLOSURE_COMPILER_ARGUMENTS = ''
    YUI_BINARY = 'java -jar yuicompressor.jar'
    YUI_CSS_ARGUMENTS = ''
    YUI_JS_ARGUMENTS = ''
    YUGLIFY_BINARY = 'yuglify'
    YUGLIFY_CSS_ARGUMENTS = '--terminal'
    YUGLIFY_JS_ARGUMENTS = '--terminal'
    CLEAN_CSS_BINARY = 'cleancss'
    CLEAN_CSS_ARGUMENTS = ''
    DATA_URI_MAX_SIZE = 1024

    # the cache backend to use
    CACHE_BACKEND = None
    # the dotted path to the function that creates the cache key
    CACHE_KEY_FUNCTION = 'compressor.cache.simple_cachekey'
    # rebuilds the cache every 30 days if nothing has changed.
    REBUILD_TIMEOUT = 60 * 60 * 24 * 30  # 30 days
    # the upper bound on how long any compression should take to be generated
    # (used against dog piling, should be a lot smaller than REBUILD_TIMEOUT
    MINT_DELAY = 30  # seconds
    # check for file changes only after a delay
    MTIME_DELAY = 10  # seconds
    # enables the offline cache -- also filled by the compress command
    OFFLINE = False
    # invalidates the offline cache after one year
    OFFLINE_TIMEOUT = 60 * 60 * 24 * 365  # 1 year
    # The context to be used when compressing the files "offline"
    OFFLINE_CONTEXT = {}
    # The name of the manifest file (e.g. filename.ext)
    OFFLINE_MANIFEST = 'manifest.json'
    # The Context to be used when TemplateFilter is used
    TEMPLATE_FILTER_CONTEXT = {}
    # Placeholder to be used instead of settings.COMPRESS_URL during offline compression.
    # Affects manifest file contents only.
    URL_PLACEHOLDER = '/__compressor_url_placeholder__/'

    # Returns the Jinja2 environment to use in offline compression.
    def JINJA2_GET_ENVIRONMENT():
        alias = 'jinja2'
        try:
            from django.template import engines
            return engines[alias].env
        except InvalidTemplateEngineError:
            raise InvalidTemplateEngineError(
                "Could not find config for '{}' "
                "in settings.TEMPLATES. "
                "COMPRESS_JINJA2_GET_ENVIRONMENT() may "
                "need to be defined in settings".format(alias))
        except ImportError:
            return None

    class Meta:
        prefix = 'compress'

    def configure_root(self, value):
        # Uses Django's STATIC_ROOT by default
        if value is None:
            value = settings.STATIC_ROOT
        if value is None:
            raise ImproperlyConfigured('COMPRESS_ROOT defaults to '
                                       + 'STATIC_ROOT, please define either')
        return os.path.normcase(os.path.abspath(value))

    def configure_url(self, value):
        # Uses Django's STATIC_URL by default
        if value is None:
            value = settings.STATIC_URL
        if not value.endswith('/'):
            raise ImproperlyConfigured("URL settings (e.g. COMPRESS_URL) "
                                       "must have a trailing slash")
        return value

    def configure_cache_backend(self, value):
        if value is None:
            value = 'default'
        return value

    def configure_offline_context(self, value):
        if not value:
            value = {'STATIC_URL': settings.STATIC_URL}
        return value

    def configure_template_filter_context(self, value):
        if not value:
            value = {'STATIC_URL': settings.STATIC_URL}
        return value

    def configure_precompilers(self, value):
        if not isinstance(value, (list, tuple)):
            raise ImproperlyConfigured("The COMPRESS_PRECOMPILERS setting "
                                       "must be a list or tuple. Check for "
                                       "missing commas.")
        return value

    def configure(self):
        data = self.configured_data
        for kind in {'css', 'js'}:
            setting_name = '%s_FILTERS' % kind.upper()
            filters = data.pop(setting_name)
            if filters is not None:
                # filters for this kind are set using <kind>_FILTERS
                if kind in data['FILTERS']:
                    raise ImproperlyConfigured(
                        "The setting {kind_setting} "
                        "conflicts with {main_setting}['{kind}']. "
                        "Remove either setting and update the other to "
                        "the correct list of filters for {kind} resources"
                        ", we recommend you keep the latter."
                        .format(
                            kind_setting=self._meta.prefixed_name(setting_name),
                            main_setting=self._meta.prefixed_name('FILTERS'),
                            kind=kind))
                data['FILTERS'][kind] = filters
            elif kind not in data['FILTERS']:
                # filters are not defined
                data['FILTERS'][kind] = default_filters[kind]
        return data



from __future__ import absolute_import
import io

import jinja2
import jinja2.ext
from jinja2 import nodes
from jinja2.ext import Extension
from jinja2.nodes import CallBlock, Call, ExtensionAttribute

from compressor.exceptions import TemplateSyntaxError, TemplateDoesNotExist


def flatten_context(context):
    if hasattr(context, 'dicts'):
        context_dict = {}

        for d in context.dicts:
            context_dict.update(d)

        return context_dict

    return context


class SpacelessExtension(Extension):
    """
    Functional "spaceless" extension equivalent to Django's.

    See: https://github.com/django/django/blob/master/django/template/defaulttags.py
    """

    tags = set(['spaceless'])

    def parse(self, parser):
        lineno = next(parser.stream).lineno
        body = parser.parse_statements(['name:endspaceless'], drop_needle=True)

        return nodes.CallBlock(self.call_method('_spaceless', []),
                               [], [], body).set_lineno(lineno)

    def _spaceless(self, caller):
        from django.utils.html import strip_spaces_between_tags

        return strip_spaces_between_tags(caller().strip())


def url_for(mod, filename):
    """
    Incomplete emulation of Flask's url_for.
    """
    try:
        from django.contrib.staticfiles.templatetags import staticfiles
    except ImportError:
        # Django 3.0+
        import django.templatetags.static as staticfiles

    if mod == "static":
        return staticfiles.static(filename)

    return ""


class Jinja2Parser(object):
    COMPRESSOR_ID = 'compressor.contrib.jinja2ext.CompressorExtension'

    def __init__(self, charset, env):
        self.charset = charset
        self.env = env

    def parse(self, template_name):
        with io.open(template_name, mode='rb') as file:
            try:
                template = self.env.parse(file.read().decode(self.charset))
            except jinja2.TemplateSyntaxError as e:
                raise TemplateSyntaxError(str(e))
            except jinja2.TemplateNotFound as e:
                raise TemplateDoesNotExist(str(e))

        return template

    def process_template(self, template, context):
        return True

    def get_init_context(self, offline_context):
        # Don't need to add filters and tests to the context, as Jinja2 will
        # automatically look for them in self.env.filters and self.env.tests.
        # This is tested by test_complex and test_templatetag.

        # Allow offline context to override the globals.
        context = self.env.globals.copy()
        context.update(flatten_context(offline_context))

        return context

    def process_node(self, template, context, node):
        pass

    def _render_nodes(self, template, context, nodes):
        compiled_node = self.env.compile(jinja2.nodes.Template(nodes))
        template = jinja2.Template.from_code(self.env, compiled_node, {})
        flat_context = flatten_context(context)

        return template.render(flat_context)

    def render_nodelist(self, template, context, node):
        return self._render_nodes(template, context, node.body)

    def render_node(self, template, context, node):
        return self._render_nodes(template, context, [node])

    def get_nodelist(self, node):
        body = getattr(node, "body", getattr(node, "nodes", []))

        if isinstance(node, jinja2.nodes.If):
            return body + node.else_

        return body

    def walk_nodes(self, node, block_name=None, context=None):
        for node in self.get_nodelist(node):
            if (isinstance(node, CallBlock)
              and isinstance(node.call, Call)
              and isinstance(node.call.node, ExtensionAttribute)
              and node.call.node.identifier == self.COMPRESSOR_ID):
                node.call.node.name = '_compress_forced'
                yield node
            else:
                for node in self.walk_nodes(node, block_name=block_name):
                    yield node



from django.test import TestCase
from django.test.utils import override_settings
from django.conf import settings
import django.contrib.staticfiles.finders
import django

import compressor.utils.staticfiles
from compressor.exceptions import FilterError
from compressor.utils import get_class

from imp import reload


def get_apps_without_staticfiles(apps):
    return [x for x in apps if x != 'django.contrib.staticfiles']


def get_apps_with_staticfiles_using_appconfig(apps):
    return get_apps_without_staticfiles(apps) + [
        'django.contrib.staticfiles.apps.StaticFilesConfig',
    ]


class StaticFilesTestCase(TestCase):

    def test_has_finders_from_staticfiles(self):
        self.assertTrue(compressor.utils.staticfiles.finders is
                        django.contrib.staticfiles.finders)

    def test_has_finders_from_staticfiles_if_configured_per_appconfig(self):
        apps = get_apps_with_staticfiles_using_appconfig(
            settings.INSTALLED_APPS)
        try:
            with override_settings(INSTALLED_APPS=apps):
                reload(compressor.utils.staticfiles)
                self.assertTrue(compressor.utils.staticfiles.finders is
                                django.contrib.staticfiles.finders)
        finally:
            reload(compressor.utils.staticfiles)

    def test_finders_is_none_if_staticfiles_is_not_installed(self):
        apps = get_apps_without_staticfiles(settings.INSTALLED_APPS)
        try:
            with override_settings(INSTALLED_APPS=apps):
                reload(compressor.utils.staticfiles)
                self.assertTrue(compressor.utils.staticfiles.finders is None)
        finally:
            reload(compressor.utils.staticfiles)


class TestGetClass(TestCase):

    def test_get_class_import_exception(self):
        with self.assertRaises(FilterError) as context:
            get_class('common.uglify.JsUglifySourcemapCompressor')

        self.assertTrue(('Failed to import common.uglify.JsUglifySourcemapCompressor. '
                         'ImportError is: No module named' in str(context.exception)))



from compressor.conf import settings
from compressor.filters import CompilerFilter


class CleanCSSFilter(CompilerFilter):
    command = "{binary} {args} -o {outfile} {infile}"
    options = (
        ("binary", settings.COMPRESS_CLEAN_CSS_BINARY),
        ("args", settings.COMPRESS_CLEAN_CSS_ARGUMENTS),
    )



from __future__ import with_statement, unicode_literals
import copy
from contextlib import contextmanager

import io
import os
from importlib import import_module

from mock import patch
from unittest import SkipTest

import six
from django.core.management import call_command
from django.core.management.base import CommandError
from django.template import Template, Context
from django.test import TestCase
from django.test.utils import override_settings

from compressor.cache import flush_offline_manifest, get_offline_manifest
from compressor.conf import settings
from compressor.exceptions import OfflineGenerationError
from compressor.management.commands.compress import Command as CompressCommand
from compressor.storage import default_storage
from compressor.utils import get_mod_func

from django.urls import get_script_prefix, set_script_prefix


def offline_context_generator():
    for i in range(1, 4):
        yield {'content': 'OK %d!' % i}


def static_url_context_generator():
    yield {'STATIC_URL': settings.STATIC_URL}


class LazyScriptNamePrefixedUrl(six.text_type):
    """
    Lazy URL with ``SCRIPT_NAME`` WSGI param as path prefix.

    .. code-block :: python

        settings.STATIC_URL = LazyScriptNamePrefixedUrl('/static/')

        # HTTP request to '/some/page/' without SCRIPT_NAME
        str(settings.STATIC_URL) == '/static/'

        # HTTP request to '/app/prefix/some/page/` with SCRIPT_NAME = '/app/prefix/'
        str(settings.STATIC_URL) == '/app/prefix/static/'

        # HTTP request to '/another/prefix/some/page/` with SCRIPT_NAME = '/another/prefix/'
        str(settings.STATIC_URL) == '/another/prefix/static/'

    The implementation is incomplete, all ``str`` methods must be overridden
    in order to work correctly with the rest of Django core.
    """
    def __str__(self):
        return get_script_prefix() + self[1:] if self.startswith('/') else self

    def __unicode__(self):
        return str(self)

    def split(self, *args, **kwargs):
        """
        Override ``.split()`` method to make it work with ``{% static %}``.
        """
        return six.text_type(self).split(*args, **kwargs)


@contextmanager
def script_prefix(new_prefix):
    """
    Override ``SCRIPT_NAME`` WSGI param, yield, then restore its original value.

    :param new_prefix: New ``SCRIPT_NAME`` value.
    """
    old_prefix = get_script_prefix()
    set_script_prefix(new_prefix)
    yield
    set_script_prefix(old_prefix)


class OfflineTestCaseMixin(object):
    CHARSET = 'utf-8'
    template_name = 'test_compressor_offline.html'
    # Change this for each test class
    templates_dir = ''
    expected_basename = 'output'
    expected_hash = ''
    # Engines to test
    engines = ('django', 'jinja2')
    additional_test_settings = None

    def setUp(self):
        # Reset template dirs, because it enables us to force compress to
        # consider only a specific directory (helps us make true,
        # independent unit tests).
        # Specify both Jinja2 and Django template locations. When the wrong
        # engine is used to parse a template, the TemplateSyntaxError will
        # cause the template to be skipped over.
        # We've hardcoded TEMPLATES[0] to be Django templates backend and
        # TEMPLATES[1] to be Jinja2 templates backend in test_settings.
        TEMPLATES = copy.deepcopy(settings.TEMPLATES)

        django_template_dir = os.path.join(
            TEMPLATES[0]['DIRS'][0], self.templates_dir)
        jinja2_template_dir = os.path.join(
            TEMPLATES[1]['DIRS'][0], self.templates_dir)

        TEMPLATES[0]['DIRS'] = [django_template_dir]
        TEMPLATES[1]['DIRS'] = [jinja2_template_dir]

        override_settings = {
            'TEMPLATES': TEMPLATES,
            'COMPRESS_ENABLED': True,
            'COMPRESS_OFFLINE': True
        }

        if 'jinja2' in self.engines:
            override_settings['COMPRESS_JINJA2_GET_ENVIRONMENT'] = (
                lambda: self._get_jinja2_env())

        if self.additional_test_settings is not None:
            override_settings.update(self.additional_test_settings)

        self.override_settings = self.settings(**override_settings)
        self.override_settings.__enter__()

        if 'django' in self.engines:
            self.template_path = os.path.join(
                django_template_dir, self.template_name)

            with io.open(self.template_path,
                         encoding=self.CHARSET) as file_:
                self.template = Template(file_.read())

        if 'jinja2' in self.engines:
            self.template_path_jinja2 = os.path.join(
                jinja2_template_dir, self.template_name)
            jinja2_env = override_settings['COMPRESS_JINJA2_GET_ENVIRONMENT']()

            with io.open(self.template_path_jinja2,
                         encoding=self.CHARSET) as file_:
                self.template_jinja2 = jinja2_env.from_string(file_.read())

    def tearDown(self):
        self.override_settings.__exit__(None, None, None)

        manifest_path = os.path.join('CACHE', 'manifest.json')
        if default_storage.exists(manifest_path):
            default_storage.delete(manifest_path)

    def _prepare_contexts(self, engine):
        contexts = settings.COMPRESS_OFFLINE_CONTEXT
        if not isinstance(contexts, (list, tuple)):
            contexts = [contexts]
        if engine == 'django':
            return [Context(c) for c in contexts]
        if engine == 'jinja2':
            return contexts
        return None

    def _render_template(self, engine):
        contexts = self._prepare_contexts(engine)
        if engine == 'django':
            return ''.join(self.template.render(c) for c in contexts)
        if engine == 'jinja2':
            return '\n'.join(
                self.template_jinja2.render(c) for c in contexts) + '\n'
        return None

    def _render_script(self, hash):
        return (
            '<script src="{}CACHE/js/{}.{}.js">'
            '</script>'.format(
                settings.COMPRESS_URL_PLACEHOLDER, self.expected_basename, hash
            )
        )

    def _render_link(self, hash):
        return (
            '<link rel="stylesheet" href="{}CACHE/css/{}.{}.css" '
            'type="text/css">'.format(
                settings.COMPRESS_URL_PLACEHOLDER, self.expected_basename, hash
            )
        )

    def _render_result(self, result, separator='\n'):
        return (separator.join(result) + '\n').replace(
            settings.COMPRESS_URL_PLACEHOLDER, six.text_type(settings.COMPRESS_URL)
        )

    def _test_offline(self, engine):
        hashes = self.expected_hash
        if not isinstance(hashes, (list, tuple)):
            hashes = [hashes]
        count, result = CompressCommand().handle_inner(engines=[engine], verbosity=0)
        self.assertEqual(len(hashes), count)
        self.assertEqual([self._render_script(h) for h in hashes], result)
        rendered_template = self._render_template(engine)
        self.assertEqual(rendered_template, self._render_result(result))

    def test_offline_django(self):
        if 'django' not in self.engines:
            raise SkipTest('This test class does not support django engine.')
        self._test_offline(engine='django')

    def test_offline_jinja2(self):
        if 'jinja2' not in self.engines:
            raise SkipTest('This test class does not support jinja2 engine.')
        self._test_offline(engine='jinja2')

    def _get_jinja2_env(self):
        import jinja2
        import jinja2.ext
        from compressor.offline.jinja2 import url_for, SpacelessExtension
        from compressor.contrib.jinja2ext import CompressorExtension

        # Extensions needed for the test cases only.
        extensions = [
            CompressorExtension,
            SpacelessExtension,
            jinja2.ext.with_,
            jinja2.ext.do,
        ]
        loader = self._get_jinja2_loader()
        env = jinja2.Environment(extensions=extensions, loader=loader)
        env.globals['url_for'] = url_for

        return env

    def _get_jinja2_loader(self):
        import jinja2

        loader = jinja2.FileSystemLoader(
            settings.TEMPLATES[1]['DIRS'], encoding=self.CHARSET)
        return loader


class OfflineCompressBasicTestCase(OfflineTestCaseMixin, TestCase):
    templates_dir = 'basic'
    expected_hash = '822ac7501287'

    @patch.object(CompressCommand, 'compress')
    def test_handle_no_args(self, compress_mock):
        compress_mock.return_value = {}, 1, []
        CompressCommand().handle()
        self.assertEqual(compress_mock.call_count, 1)

    @patch.object(CompressCommand, 'compress')
    def test_handle_compress_disabled(self, compress_mock):
        with self.settings(COMPRESS_ENABLED=False):
            with self.assertRaises(CommandError):
                CompressCommand().handle()
        self.assertEqual(compress_mock.call_count, 0)

    @patch.object(CompressCommand, 'compress')
    def test_handle_compress_offline_disabled(self, compress_mock):
        with self.settings(COMPRESS_OFFLINE=False):
            with self.assertRaises(CommandError):
                CompressCommand().handle()
        self.assertEqual(compress_mock.call_count, 0)

    @patch.object(CompressCommand, 'compress')
    def test_handle_compress_offline_disabled_force(self, compress_mock):
        compress_mock.return_value = {}, 1, []
        with self.settings(COMPRESS_OFFLINE=False):
            CompressCommand().handle(force=True)
        self.assertEqual(compress_mock.call_count, 1)

    def test_rendering_without_manifest_raises_exception(self):
        # flush cached manifest
        flush_offline_manifest()
        self.assertRaises(OfflineGenerationError,
                          self.template.render, Context({}))

    def test_rendering_without_manifest_raises_exception_jinja2(self):
        # flush cached manifest
        flush_offline_manifest()
        self.assertRaises(OfflineGenerationError,
                          self.template_jinja2.render, {})

    def _test_deleting_manifest_does_not_affect_rendering(self, engine):
        count, result = CompressCommand().handle_inner(engines=[engine], verbosity=0)
        get_offline_manifest()
        manifest_path = os.path.join('CACHE', 'manifest.json')
        if default_storage.exists(manifest_path):
            default_storage.delete(manifest_path)
        self.assertEqual(1, count)
        self.assertEqual([self._render_script(self.expected_hash)], result)
        rendered_template = self._render_template(engine)
        self.assertEqual(rendered_template, self._render_result(result))

    def test_deleting_manifest_does_not_affect_rendering(self):
        for engine in self.engines:
            self._test_deleting_manifest_does_not_affect_rendering(engine)

    def test_get_loaders(self):
        TEMPLATE_LOADERS = (
            ('django.template.loaders.cached.Loader', (
                'django.template.loaders.filesystem.Loader',
                'django.template.loaders.app_directories.Loader',
            )),
        )
        with self.settings(TEMPLATE_LOADERS=TEMPLATE_LOADERS):
            from django.template.loaders.filesystem import (
                Loader as FileSystemLoader)
            from django.template.loaders.app_directories import (
                Loader as AppDirectoriesLoader)
            loaders = CompressCommand().get_loaders()
            self.assertTrue(isinstance(loaders[0], FileSystemLoader))
            self.assertTrue(isinstance(loaders[1], AppDirectoriesLoader))

    @patch("compressor.offline.django.DjangoParser.render_node",
           side_effect=Exception(b"non-ascii character here:\xc3\xa4"))
    def test_non_ascii_exception_messages(self, mock):
        with self.assertRaises(CommandError):
            CompressCommand().handle(verbosity=0)


class OfflineCompressSkipDuplicatesTestCase(OfflineTestCaseMixin, TestCase):
    templates_dir = 'test_duplicate'

    def _test_offline(self, engine):
        count, result = CompressCommand().handle_inner(engines=[engine], verbosity=0)
        # Only one block compressed, the second identical one was skipped.
        self.assertEqual(1, count)
        # Only 1 <script> block in returned result as well.
        self.assertEqual([self._render_script('822ac7501287')], result)
        rendered_template = self._render_template(engine)
        # But rendering the template returns both (identical) scripts.
        self.assertEqual(
            rendered_template, self._render_result(result * 2, ''))


class SuperMixin:
    # Block.super not supported for Jinja2 yet.
    engines = ('django',)


class OfflineCompressBlockSuperTestCase(
        SuperMixin, OfflineTestCaseMixin, TestCase):
    templates_dir = 'test_block_super'
    expected_hash = '817b5defb197'


class OfflineCompressBlockSuperMultipleTestCase(
        SuperMixin, OfflineTestCaseMixin, TestCase):
    templates_dir = 'test_block_super_multiple'
    expected_hash = 'd3f749e83c81'


class OfflineCompressBlockSuperMultipleCachedLoaderTestCase(
        SuperMixin, OfflineTestCaseMixin, TestCase):
    templates_dir = 'test_block_super_multiple_cached'
    expected_hash = '055f88f4751f'
    additional_test_settings = {
        'TEMPLATE_LOADERS': (
            ('django.template.loaders.cached.Loader', (
                'django.template.loaders.filesystem.Loader',
                'django.template.loaders.app_directories.Loader',
            )),
        )
    }


class OfflineCompressBlockSuperTestCaseWithExtraContent(
        SuperMixin, OfflineTestCaseMixin, TestCase):
    templates_dir = 'test_block_super_extra'

    def _test_offline(self, engine):
        count, result = CompressCommand().handle_inner(engines=[engine], verbosity=0)
        self.assertEqual(2, count)
        self.assertEqual([
            self._render_script('bfcec76e0f28'),
            self._render_script('817b5defb197')
        ], result)
        rendered_template = self._render_template(engine)
        self.assertEqual(rendered_template, self._render_result(result, ''))


class OfflineCompressConditionTestCase(OfflineTestCaseMixin, TestCase):
    templates_dir = 'test_condition'
    expected_hash = 'a3275743dc69'
    additional_test_settings = {
        'COMPRESS_OFFLINE_CONTEXT': {
            'condition': 'red',
        }
    }


class OfflineCompressTemplateTagTestCase(OfflineTestCaseMixin, TestCase):
    templates_dir = 'test_templatetag'
    expected_hash = '2bb88185b4f5'


class OfflineCompressStaticTemplateTagTestCase(OfflineTestCaseMixin, TestCase):
    templates_dir = 'test_static_templatetag'
    expected_hash = 'be0b1eade28b'


class OfflineCompressTemplateTagNamedTestCase(OfflineTestCaseMixin, TestCase):
    templates_dir = 'test_templatetag_named'
    expected_basename = 'output_name'
    expected_hash = '822ac7501287'


class OfflineCompressTestCaseWithContext(OfflineTestCaseMixin, TestCase):
    templates_dir = 'test_with_context'
    expected_hash = 'c6bf81bca7ad'
    additional_test_settings = {
        'COMPRESS_OFFLINE_CONTEXT': {
            'content': 'OK!',
        }
    }


class OfflineCompressTestCaseWithContextSuper(
        SuperMixin, OfflineTestCaseMixin, TestCase):
    templates_dir = 'test_with_context_super'
    expected_hash = 'dd79e1bd1527'
    additional_test_settings = {
        'COMPRESS_OFFLINE_CONTEXT': {
            'content': 'OK!',
        }
    }


class OfflineCompressTestCaseWithContextList(OfflineTestCaseMixin, TestCase):
    templates_dir = 'test_with_context'
    expected_hash = ['8b4a7452e1c5', '55b3123e884c', 'bfc63829cc58']
    additional_test_settings = {
        'COMPRESS_OFFLINE_CONTEXT': list(offline_context_generator())
    }

    def _prepare_contexts(self, engine):
        if engine == 'django':
            return [Context(c) for c in settings.COMPRESS_OFFLINE_CONTEXT]
        if engine == 'jinja2':
            return settings.COMPRESS_OFFLINE_CONTEXT
        return None


class OfflineCompressTestCaseWithContextListSuper(
        SuperMixin, OfflineCompressTestCaseWithContextList):
    templates_dir = 'test_with_context_super'
    expected_hash = ['b39975a8f6ea', 'ed565a1d262f', '6ac9e4b29feb']
    additional_test_settings = {
        'COMPRESS_OFFLINE_CONTEXT': list(offline_context_generator())
    }


class OfflineCompressTestCaseWithContextGenerator(
        OfflineTestCaseMixin, TestCase):
    templates_dir = 'test_with_context'
    expected_hash = ['8b4a7452e1c5', '55b3123e884c', 'bfc63829cc58']
    additional_test_settings = {
        'COMPRESS_OFFLINE_CONTEXT': 'compressor.tests.test_offline.'
                                    'offline_context_generator'
    }

    def _prepare_contexts(self, engine):
        module, function = get_mod_func(settings.COMPRESS_OFFLINE_CONTEXT)
        contexts = getattr(import_module(module), function)()
        if engine == 'django':
            return (Context(c) for c in contexts)
        if engine == 'jinja2':
            return contexts
        return None


class OfflineCompressTestCaseWithContextGeneratorSuper(
        SuperMixin, OfflineCompressTestCaseWithContextGenerator):
    templates_dir = 'test_with_context_super'
    expected_hash = ['b39975a8f6ea', 'ed565a1d262f', '6ac9e4b29feb']
    additional_test_settings = {
        'COMPRESS_OFFLINE_CONTEXT': 'compressor.tests.test_offline.'
                                    'offline_context_generator'
    }


class OfflineCompressStaticUrlIndependenceTestCase(
        OfflineCompressTestCaseWithContextGenerator):
    """
    Test that the offline manifest is independent of STATIC_URL.
    I.e. users can use the manifest with any other STATIC_URL in the future.
    """
    templates_dir = 'test_static_url_independence'
    expected_hash = 'b0bfc3754fd4'
    additional_test_settings = {
        'STATIC_URL': '/custom/static/url/',
        # We use ``COMPRESS_OFFLINE_CONTEXT`` generator to make sure that
        # ``STATIC_URL`` is not cached when rendering the template.
        'COMPRESS_OFFLINE_CONTEXT': (
            'compressor.tests.test_offline.static_url_context_generator'
        )
    }

    def _test_offline(self, engine):
        count, result = CompressCommand().handle_inner(engines=[engine], verbosity=0)
        self.assertEqual(1, count)
        self.assertEqual([self._render_script(self.expected_hash)], result)
        self.assertEqual(
            self._render_template(engine), self._render_result(result))

        # Changing STATIC_URL setting doesn't break things despite that
        # offline compression was made with different STATIC_URL.
        with self.settings(STATIC_URL='/another/static/url/'):
            self.assertEqual(
                self._render_template(engine), self._render_result(result))


class OfflineCompressTestCaseWithContextVariableInheritance(
        OfflineTestCaseMixin, TestCase):
    templates_dir = 'test_with_context_variable_inheritance'
    expected_hash = 'b8376aad1357'
    additional_test_settings = {
        'COMPRESS_OFFLINE_CONTEXT': {
            'parent_template': 'base.html',
        }
    }

    def _render_result(self, result, separator='\n'):
        return '\n' + super(
            OfflineCompressTestCaseWithContextVariableInheritance, self
        )._render_result(result, separator)


class OfflineCompressTestCaseWithContextVariableInheritanceSuper(
        SuperMixin, OfflineTestCaseMixin, TestCase):
    templates_dir = 'test_with_context_variable_inheritance_super'
    additional_test_settings = {
        'COMPRESS_OFFLINE_CONTEXT': [{
            'parent_template': 'base1.html',
        }, {
            'parent_template': 'base2.html',
        }]
    }
    expected_hash = ['cee48db7cedc', 'c877c436363a']


class OfflineCompressTestCaseWithContextGeneratorImportError(
        OfflineTestCaseMixin, TestCase):
    templates_dir = 'test_with_context'

    def _test_offline(self, engine):
        # Test that we are properly generating ImportError when
        # COMPRESS_OFFLINE_CONTEXT looks like a function but can't be imported
        # for whatever reason.

        with self.settings(
                COMPRESS_OFFLINE_CONTEXT='invalid_mod.invalid_func'):
            # Path with invalid module name -- ImportError:
            self.assertRaises(
                ImportError, CompressCommand().handle_inner, engines=[engine])

        with self.settings(COMPRESS_OFFLINE_CONTEXT='compressor'):
            # Valid module name only without function -- AttributeError:
            self.assertRaises(
                ImportError, CompressCommand().handle_inner, engines=[engine])

        with self.settings(
                COMPRESS_OFFLINE_CONTEXT='compressor.tests.invalid_function'):
            # Path with invalid function name -- AttributeError:
            self.assertRaises(
                ImportError, CompressCommand().handle_inner, engines=[engine])

        with self.settings(
                COMPRESS_OFFLINE_CONTEXT='compressor.tests.test_offline'):
            # Path without function attempts call on module -- TypeError:
            self.assertRaises(
                ImportError, CompressCommand().handle_inner, engines=[engine])

        valid_path = 'compressor.tests.test_offline.offline_context_generator'
        with self.settings(COMPRESS_OFFLINE_CONTEXT=valid_path):
            # Valid path to generator function -- no ImportError:

            try:
                CompressCommand().handle_inner(engines=[engine], verbosity=0)
            except ImportError:
                self.fail('Valid path to offline context generator must'
                          ' not raise ImportError.')


class OfflineCompressTestCaseErrors(OfflineTestCaseMixin, TestCase):
    templates_dir = 'test_error_handling'

    def _test_offline(self, engine):
        count, result = CompressCommand().handle_inner(engines=[engine], verbosity=0)

        if engine == 'django':
            self.assertEqual(2, count)
        else:
            # Because we use env.parse in Jinja2Parser, the engine does not
            # actually load the 'extends' and 'includes' templates, and so
            # it is unable to detect that they are missing. So all the
            # 'compress' nodes are processed correctly.
            self.assertEqual(4, count)
            self.assertEqual(engine, 'jinja2')
            self.assertIn(self._render_link('7ff52cb38987'), result)
            self.assertIn(self._render_link('2db2b4d36380'), result)

        self.assertIn(self._render_script('eeabdac29232'), result)
        self.assertIn(self._render_script('9a7f06880ce3'), result)


class OfflineCompressTestCaseWithError(OfflineTestCaseMixin, TestCase):
    templates_dir = 'test_error_handling'
    additional_test_settings = {
        'COMPRESS_PRECOMPILERS': (('text/coffeescript', 'nonexisting-binary'),)
    }

    def _test_offline(self, engine):
        """
        Test that a CommandError is raised with DEBUG being False as well as
        True, as otherwise errors in configuration will never show in
        production.
        """
        with self.settings(DEBUG=True):
            self.assertRaises(
                CommandError, CompressCommand().handle_inner, engines=[engine], verbosity=0)

        with self.settings(DEBUG=False):
            self.assertRaises(
                CommandError, CompressCommand().handle_inner, engines=[engine], verbosity=0)


class OfflineCompressEmptyTag(OfflineTestCaseMixin, TestCase):
    """
        In case of a compress template tag with no content, an entry
        will be added to the manifest with an empty string as value.
        This test makes sure there is no recompression happening when
        compressor encounters such an emptystring in the manifest.
    """
    templates_dir = 'basic'
    expected_hash = '822ac7501287'

    def _test_offline(self, engine):
        CompressCommand().handle_inner(engines=[engine], verbosity=0)
        manifest = get_offline_manifest()
        manifest[list(manifest)[0]] = ''
        self.assertEqual(self._render_template(engine), '\n')


class OfflineCompressBlockSuperBaseCompressed(OfflineTestCaseMixin, TestCase):
    template_names = ['base.html', 'base2.html',
                      'test_compressor_offline.html']
    templates_dir = 'test_block_super_base_compressed'
    expected_hash_offline = ['e4e9263fa4c0', '9cecd41a505f', 'd3f749e83c81']
    expected_hash = ['028c3fc42232', '2e9d3f5545a6', 'd3f749e83c81']
    # Block.super not supported for Jinja2 yet.
    engines = ('django',)

    def setUp(self):
        super(OfflineCompressBlockSuperBaseCompressed, self).setUp()

        self.template_paths = []
        self.templates = []
        for template_name in self.template_names:
            template_path = os.path.join(
                settings.TEMPLATES[0]['DIRS'][0], template_name)
            self.template_paths.append(template_path)
            with io.open(template_path,
                         encoding=self.CHARSET) as file_:
                template = Template(file_.read())
            self.templates.append(template)

    def _render_template(self, template, engine):
        if engine == 'django':
            return template.render(Context(settings.COMPRESS_OFFLINE_CONTEXT))
        elif engine == 'jinja2':
            return template.render(settings.COMPRESS_OFFLINE_CONTEXT) + '\n'
        else:
            return None

    def _test_offline(self, engine):
        count, result = CompressCommand().handle_inner(engines=[engine], verbosity=0)
        self.assertEqual(len(self.expected_hash), count)
        for expected_hash, template in zip(self.expected_hash_offline, self.templates):
            expected = self._render_script(expected_hash)
            self.assertIn(expected, result)
            rendered_template = self._render_template(template, engine)
            self.assertEqual(
                rendered_template, self._render_result([expected]))


class OfflineCompressInlineNonAsciiTestCase(OfflineTestCaseMixin, TestCase):
    templates_dir = 'test_inline_non_ascii'
    additional_test_settings = {
        'COMPRESS_OFFLINE_CONTEXT': {
            'test_non_ascii_value': '\u2014',
        }
    }

    def _test_offline(self, engine):
        _, result = CompressCommand().handle_inner(engines=[engine], verbosity=0)
        rendered_template = self._render_template(engine)
        self.assertEqual(rendered_template, ''.join(result) + '\n')


class OfflineCompressComplexTestCase(OfflineTestCaseMixin, TestCase):
    templates_dir = 'test_complex'
    additional_test_settings = {
        'COMPRESS_OFFLINE_CONTEXT': {
            'condition': 'OK!',
            # Django templating does not allow definition of tuples in the
            # templates.
            # Make sure this is same as test_templates_jinja2/test_complex.
            'my_names': ('js/one.js', 'js/nonasc.js'),
        }
    }

    def _test_offline(self, engine):
        count, result = CompressCommand().handle_inner(engines=[engine], verbosity=0)
        self.assertEqual(3, count)
        self.assertEqual([
            self._render_script('76a82cfab9ab'),
            self._render_script('7219642b8ab4'),
            self._render_script('567bb77b13db')
        ], result)
        rendered_template = self._render_template(engine)
        self.assertEqual(
            rendered_template, self._render_result([result[0], result[2]], ''))


class OfflineCompressExtendsRecursionTestCase(OfflineTestCaseMixin, TestCase):
    """
    Test that templates extending templates with the same name
    (e.g. admin/index.html) don't cause an infinite test_extends_recursion
    """
    templates_dir = 'test_extends_recursion'

    INSTALLED_APPS = [
        'django.contrib.admin',
        'django.contrib.auth',
        'django.contrib.contenttypes',
        'django.contrib.staticfiles',
        'compressor',
    ]

    @override_settings(INSTALLED_APPS=INSTALLED_APPS)
    def _test_offline(self, engine):
        count, _ = CompressCommand().handle_inner(engines=[engine], verbosity=0)
        self.assertEqual(count, 1)


class TestCompressCommand(OfflineTestCaseMixin, TestCase):
    templates_dir = "test_compress_command"

    def _test_offline(self, engine):
        raise SkipTest("Not utilized for this test case")

    def _build_expected_manifest(self, expected):
        return {
            k: self._render_script(v) for k, v in expected.items()
        }

    def test_multiple_engines(self):
        opts = {
            "force": True,
            "verbosity": 0,
        }

        call_command('compress', engines=["django"], **opts)
        manifest_django = get_offline_manifest()
        manifest_django_expected = self._build_expected_manifest(
            {'0fed9c02607acba22316a328075a81a74e0983ae79470daa9d3707a337623dc3': '0241107e9a9a'})
        self.assertEqual(manifest_django, manifest_django_expected)

        call_command('compress', engines=["jinja2"], **opts)
        manifest_jinja2 = get_offline_manifest()
        manifest_jinja2_expected = self._build_expected_manifest(
            {'077408d23d4a829b8f88db2eadcf902b29d71b14f94018d900f38a3f8ed24c94': '5694ca83dd14'})
        self.assertEqual(manifest_jinja2, manifest_jinja2_expected)

        call_command('compress', engines=["django", "jinja2"], **opts)
        manifest_both = get_offline_manifest()
        manifest_both_expected = self._build_expected_manifest(
            {'0fed9c02607acba22316a328075a81a74e0983ae79470daa9d3707a337623dc3': '0241107e9a9a',
             '077408d23d4a829b8f88db2eadcf902b29d71b14f94018d900f38a3f8ed24c94': '5694ca83dd14'})
        self.assertEqual(manifest_both, manifest_both_expected)


class OfflineCompressTestCaseWithLazyStringAlikeUrls(OfflineCompressTestCaseWithContextGenerator):
    """
    Test offline compressing with ``STATIC_URL`` and ``COMPRESS_URL`` as instances of
    *lazy string-alike objects* instead of strings.

    In particular, lazy string-alike objects that add ``SCRIPT_NAME`` WSGI param
    as URL path prefix.

    For example:

    - We've generated offline assets and deployed them with our Django project.
    - We've configured HTTP server (e.g. Nginx) to serve our app at two different URLs:
      ``http://example.com/my/app/`` and ``http://app.example.com/``.
    - Both URLs are leading to the same app, but in the first case we pass
      ``SCRIPT_NAME = /my/app/`` to WSGI app server (e.g. to uWSGI, which is *behind* Nginx).
    - Django (1.11.7, as of today) *ignores* ``SCRIPT_NAME`` when generating
      static URLs, while it uses ``SCRIPT_NAME`` when generating Django views URLs -
      see https://code.djangoproject.com/ticket/25598.
    - As a solution - we can use a lazy string-alike object instead of ``str`` for ``STATIC_URL``
      so it will know about ``SCRIPT_NAME`` and add it as a prefix every time we do any
      string operation with ``STATIC_URL``.
    - However, there are some cases when we cannot force CPython to render our lazy string
      correctly - e.g. ``some_string.replace(STATIC_URL, '...')``. So we need to do explicit
      ``str`` type cast: ``some_string.replace(str(STATIC_URL), '...')``.
    """
    templates_dir = 'test_static_templatetag'
    additional_test_settings = {
        'STATIC_URL': LazyScriptNamePrefixedUrl('/static/'),
        'COMPRESS_URL': LazyScriptNamePrefixedUrl('/static/'),
        # We use ``COMPRESS_OFFLINE_CONTEXT`` generator to make sure that
        # ``STATIC_URL`` is not cached when rendering the template.
        'COMPRESS_OFFLINE_CONTEXT': (
            'compressor.tests.test_offline.static_url_context_generator'
        )
    }
    expected_hash = 'be0b1eade28b'

    def _test_offline(self, engine):
        count, result = CompressCommand().handle_inner(engines=[engine], verbosity=0)
        self.assertEqual(1, count)

        # Change ``SCRIPT_NAME`` WSGI param - it can be changed on every HTTP request,
        # e.g. passed via HTTP header.
        for script_name in ['', '/app/prefix/', '/another/prefix/']:
            with script_prefix(script_name):
                self.assertEqual(
                    six.text_type(settings.STATIC_URL),
                    script_name.rstrip('/') + '/static/'
                )

                self.assertEqual(
                    six.text_type(settings.COMPRESS_URL),
                    script_name.rstrip('/') + '/static/'
                )

                expected_result = self._render_result(result)
                actual_result = self._render_template(engine)

                self.assertEqual(actual_result, expected_result)
                self.assertIn(six.text_type(settings.COMPRESS_URL), actual_result)



from __future__ import unicode_literals
import errno
import gzip
import os
from datetime import datetime
import time

from django.core.files.storage import FileSystemStorage, get_storage_class
from django.utils.functional import LazyObject, SimpleLazyObject

from compressor.conf import settings


class CompressorFileStorage(FileSystemStorage):
    """
    Standard file system storage for files handled by django-compressor.

    The defaults for ``location`` and ``base_url`` are ``COMPRESS_ROOT`` and
    ``COMPRESS_URL``.

    """
    def __init__(self, location=None, base_url=None, *args, **kwargs):
        if location is None:
            location = settings.COMPRESS_ROOT
        if base_url is None:
            base_url = settings.COMPRESS_URL
        super(CompressorFileStorage, self).__init__(location, base_url,
                                                    *args, **kwargs)

    def accessed_time(self, name):
        return datetime.fromtimestamp(os.path.getatime(self.path(name)))

    def created_time(self, name):
        return datetime.fromtimestamp(os.path.getctime(self.path(name)))

    def modified_time(self, name):
        return datetime.fromtimestamp(os.path.getmtime(self.path(name)))

    def get_available_name(self, name, max_length=None):
        """
        Deletes the given file if it exists.
        """
        if self.exists(name):
            self.delete(name)
        return name

    def delete(self, name):
        """
        Handle deletion race condition present in Django prior to 1.4
        https://code.djangoproject.com/ticket/16108
        """
        try:
            super(CompressorFileStorage, self).delete(name)
        except OSError as e:
            if e.errno != errno.ENOENT:
                raise


compressor_file_storage = SimpleLazyObject(
    lambda: get_storage_class('compressor.storage.CompressorFileStorage')())


class GzipCompressorFileStorage(CompressorFileStorage):
    """
    File system storage that stores gzipped files in addition to the usual files.
    """
    def save(self, filename, content):
        filename = super(GzipCompressorFileStorage, self).save(filename, content)
        orig_path = self.path(filename)
        compressed_path = '%s.gz' % orig_path

        with open(orig_path, 'rb') as f_in, open(compressed_path, 'wb') as f_out:
            with gzip.GzipFile(fileobj=f_out) as gz_out:
                gz_out.write(f_in.read())

        # Ensure the file timestamps match.
        # os.stat() returns nanosecond resolution on Linux, but os.utime()
        # only sets microsecond resolution.  Set times on both files to
        # ensure they are equal.
        stamp = time.time()
        os.utime(orig_path, (stamp, stamp))
        os.utime(compressed_path, (stamp, stamp))

        return filename


class BrotliCompressorFileStorage(CompressorFileStorage):
    """
    File system storage that stores brotli files in addition to the usual files.
    """
    chunk_size = 1024

    def save(self, filename, content):
        filename = super(BrotliCompressorFileStorage, self).save(filename, content)
        orig_path = self.path(filename)
        compressed_path = '%s.br' % orig_path

        import brotli
        br_compressor = brotli.Compressor()
        with open(orig_path, 'rb') as f_in, open(compressed_path, 'wb') as f_out:
            for f_in_data in iter(lambda: f_in.read(self.chunk_size), b''):
                compressed_data = br_compressor.process(f_in_data)
                if not compressed_data:
                    compressed_data = br_compressor.flush()
                f_out.write(compressed_data)
            f_out.write(br_compressor.finish())
        # Ensure the file timestamps match.
        # os.stat() returns nanosecond resolution on Linux, but os.utime()
        # only sets microsecond resolution.  Set times on both files to
        # ensure they are equal.
        stamp = time.time()
        os.utime(orig_path, (stamp, stamp))
        os.utime(compressed_path, (stamp, stamp))

        return filename


class DefaultStorage(LazyObject):
    def _setup(self):
        self._wrapped = get_storage_class(settings.COMPRESS_STORAGE)()


default_storage = DefaultStorage()



import os

TEST_DIR = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'tests')


CACHES = {
    'default': {
        'BACKEND': 'django.core.cache.backends.locmem.LocMemCache',
        'LOCATION': 'unique-snowflake'
    }
}

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': ':memory:',
    }
}

INSTALLED_APPS = [
    'django.contrib.staticfiles',
    'compressor',
    'sekizai',
]

STATICFILES_FINDERS = [
    'django.contrib.staticfiles.finders.FileSystemFinder',
    'django.contrib.staticfiles.finders.AppDirectoriesFinder',
    'compressor.finders.CompressorFinder',
]

STATIC_URL = '/static/'


STATIC_ROOT = os.path.join(TEST_DIR, 'static')

TEMPLATES = [{
    'BACKEND': 'django.template.backends.django.DjangoTemplates',
    'APP_DIRS': True,
    'DIRS': [
        # Specifically choose a name that will not be considered
        # by app_directories loader, to make sure each test uses
        # a specific template without considering the others.
        os.path.join(TEST_DIR, 'test_templates'),
    ],
}, {
    'BACKEND': 'django.template.backends.jinja2.Jinja2',
    'APP_DIRS': True,
    'DIRS': [
        # Specifically choose a name that will not be considered
        # by app_directories loader, to make sure each test uses
        # a specific template without considering the others.
        os.path.join(TEST_DIR, 'test_templates_jinja2'),
    ],
}]

SECRET_KEY = "iufoj=mibkpdz*%bob952x(%49rqgv8gg45k36kjcg76&-y5=!"

PASSWORD_HASHERS = (
    'django.contrib.auth.hashers.UnsaltedMD5PasswordHasher',
)

MIDDLEWARE_CLASSES = []



import fnmatch
import os

from django.core.management.base import BaseCommand, CommandError

from compressor.conf import settings
from compressor.cache import cache, get_mtime, get_mtime_cachekey


class Command(BaseCommand):
    help = "Add or remove all mtime values from the cache"

    def add_arguments(self, parser):
        parser.add_argument(
            '-i', '--ignore', action='append', default=[],
            dest='ignore_patterns', metavar='PATTERN',
            help="Ignore files or directories matching this glob-style "
                 "pattern. Use multiple times to ignore more."),
        parser.add_argument(
            '--no-default-ignore', action='store_false',
            dest='use_default_ignore_patterns', default=True,
            help="Don't ignore the common private glob-style patterns 'CVS', "
                 "'.*' and '*~'."),
        parser.add_argument(
            '--follow-links', dest='follow_links', action='store_true',
            help="Follow symlinks when traversing the COMPRESS_ROOT "
                 "(which defaults to STATIC_ROOT). Be aware that using this "
                 "can lead to infinite recursion if a link points to a parent "
                 "directory of itself."),
        parser.add_argument(
            '-c', '--clean', dest='clean', action='store_true',
            help="Remove all items"),
        parser.add_argument(
            '-a', '--add', dest='add', action='store_true',
            help="Add all items"),

    def is_ignored(self, path):
        """
        Return True or False depending on whether the ``path`` should be
        ignored (if it matches any pattern in ``ignore_patterns``).
        """
        for pattern in self.ignore_patterns:
            if fnmatch.fnmatchcase(path, pattern):
                return True
        return False

    def handle(self, **options):
        ignore_patterns = options['ignore_patterns']
        if options['use_default_ignore_patterns']:
            ignore_patterns += ['CVS', '.*', '*~']
            options['ignore_patterns'] = ignore_patterns
        self.ignore_patterns = ignore_patterns

        if ((options['add'] and options['clean'])
                or (not options['add'] and not options['clean'])):
            raise CommandError('Please specify either "--add" or "--clean"')

        if not settings.COMPRESS_MTIME_DELAY:
            raise CommandError(
                'mtime caching is currently disabled. Please '
                'set the COMPRESS_MTIME_DELAY setting to a number of seconds.')

        files_to_add = set()
        keys_to_delete = set()

        for root, dirs, files in os.walk(settings.COMPRESS_ROOT,
                                         followlinks=options['follow_links']):
            for dir_ in dirs:
                if self.is_ignored(dir_):
                    dirs.remove(dir_)
            for filename in files:
                common = "".join(root.split(settings.COMPRESS_ROOT))
                if common.startswith(os.sep):
                    common = common[len(os.sep):]
                if self.is_ignored(os.path.join(common, filename)):
                    continue
                filename = os.path.join(root, filename)
                keys_to_delete.add(get_mtime_cachekey(filename))
                if options['add']:
                    files_to_add.add(filename)

        if keys_to_delete:
            cache.delete_many(list(keys_to_delete))
            self.stdout.write("Deleted mtimes of %d files from the cache."
                              % len(keys_to_delete))

        if files_to_add:
            for filename in files_to_add:
                get_mtime(filename)
            self.stdout.write("Added mtimes of %d files to cache."
                              % len(files_to_add))



from __future__ import absolute_import
from copy import copy

from django import template
from django.template import Context
from django.template.base import Node, VariableNode, TextNode, NodeList
from django.template.defaulttags import IfNode
from django.template.loader import get_template
from django.template.loader_tags import BLOCK_CONTEXT_KEY, ExtendsNode, BlockNode, BlockContext


from compressor.exceptions import TemplateSyntaxError, TemplateDoesNotExist
from compressor.templatetags.compress import CompressorNode


def handle_extendsnode(extendsnode, context):
    """Create a copy of Node tree of a derived template replacing
    all blocks tags with the nodes of appropriate blocks.
    Also handles {{ block.super }} tags.
    """
    if BLOCK_CONTEXT_KEY not in context.render_context:
        context.render_context[BLOCK_CONTEXT_KEY] = BlockContext()
    block_context = context.render_context[BLOCK_CONTEXT_KEY]
    blocks = dict((n.name, n) for n in
                  extendsnode.nodelist.get_nodes_by_type(BlockNode))
    block_context.add_blocks(blocks)

    compiled_parent = extendsnode.get_parent(context)
    parent_nodelist = compiled_parent.nodelist
    # If the parent template has an ExtendsNode it is not the root.
    for node in parent_nodelist:
        # The ExtendsNode has to be the first non-text node.
        if not isinstance(node, TextNode):
            if isinstance(node, ExtendsNode):
                return handle_extendsnode(node, context)
            break
    # Add blocks of the root template to block context.
    blocks = dict((n.name, n) for n in
                  parent_nodelist.get_nodes_by_type(BlockNode))
    block_context.add_blocks(blocks)

    block_stack = []
    new_nodelist = remove_block_nodes(parent_nodelist, block_stack, block_context)
    return new_nodelist


def remove_block_nodes(nodelist, block_stack, block_context):
    new_nodelist = NodeList()
    for node in nodelist:
        if isinstance(node, VariableNode):
            var_name = node.filter_expression.token.strip()
            if var_name == 'block.super':
                if not block_stack:
                    continue
                node = block_context.get_block(block_stack[-1].name)
                if not node:
                    continue
        if isinstance(node, BlockNode):
            expanded_block = expand_blocknode(node, block_stack, block_context)
            new_nodelist.extend(expanded_block)
        else:
            # IfNode has nodelist as a @property so we can not modify it
            if isinstance(node, IfNode):
                node = copy(node)
                for i, (condition, sub_nodelist) in enumerate(node.conditions_nodelists):
                    sub_nodelist = remove_block_nodes(sub_nodelist, block_stack, block_context)
                    node.conditions_nodelists[i] = (condition, sub_nodelist)
            else:
                for attr in node.child_nodelists:
                    sub_nodelist = getattr(node, attr, None)
                    if sub_nodelist:
                        sub_nodelist = remove_block_nodes(sub_nodelist, block_stack, block_context)
                        node = copy(node)
                        setattr(node, attr, sub_nodelist)
            new_nodelist.append(node)
    return new_nodelist


def expand_blocknode(node, block_stack, block_context):
    popped_block = block = block_context.pop(node.name)
    if block is None:
        block = node
    block_stack.append(block)
    expanded_nodelist = remove_block_nodes(block.nodelist, block_stack, block_context)
    block_stack.pop()
    if popped_block is not None:
        block_context.push(node.name, popped_block)
    return expanded_nodelist


class DjangoParser(object):
    def __init__(self, charset):
        self.charset = charset

    def parse(self, template_name):
        try:
            return get_template(template_name).template
        except template.TemplateSyntaxError as e:
            raise TemplateSyntaxError(str(e))
        except template.TemplateDoesNotExist as e:
            raise TemplateDoesNotExist(str(e))

    def process_template(self, template, context):
        return True

    def get_init_context(self, offline_context):
        return offline_context

    def process_node(self, template, context, node):
        pass

    def render_nodelist(self, template, context, node):
        context.template = template
        return node.nodelist.render(context)

    def render_node(self, template, context, node):
        return node.render(context, forced=True)

    def get_nodelist(self, node, original, context=None):
        if isinstance(node, ExtendsNode):
            try:
                if context is None:
                    context = Context()
                context.template = original
                return handle_extendsnode(node, context)
            except template.TemplateSyntaxError as e:
                raise TemplateSyntaxError(str(e))
            except template.TemplateDoesNotExist as e:
                raise TemplateDoesNotExist(str(e))

        # Check if node is an ```if``` switch with true and false branches
        nodelist = []
        if isinstance(node, Node):
            for attr in node.child_nodelists:
                # see https://github.com/django-compressor/django-compressor/pull/825
                # and linked issues/PRs for a discussion on the `None) or []` part
                nodelist += getattr(node, attr, None) or []
        else:
            nodelist = getattr(node, 'nodelist', [])
        return nodelist

    def walk_nodes(self, node, original=None, context=None):
        if original is None:
            original = node
        for node in self.get_nodelist(node, original, context):
            if isinstance(node, CompressorNode) \
                    and node.is_offline_compression_enabled(forced=True):
                yield node
            else:
                for node in self.walk_nodes(node, original, context):
                    yield node



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
import os

from compressor.exceptions import FilterError


def get_class(class_string, exception=FilterError):
    """
    Convert a string version of a function name to the callable object.
    """
    if not hasattr(class_string, '__bases__'):
        try:
            class_string = str(class_string)
            mod_name, class_name = get_mod_func(class_string)
            if class_name:
                return getattr(__import__(mod_name, {}, {}, [str('')]), class_name)
        except AttributeError as e:
            raise exception('Failed to import %s. AttributeError is: %s' % (class_string, e))
        except ImportError as e:
            raise exception('Failed to import %s. ImportError is: %s' % (class_string, e))

        raise exception("Invalid class path '%s'" % class_string)


def get_mod_func(callback):
    """
    Converts 'django.views.news.stories.story_detail' to
    ('django.views.news.stories', 'story_detail')
    """
    try:
        dot = callback.rindex('.')
    except ValueError:
        return callback, ''
    return callback[:dot], callback[dot + 1:]


def get_pathext(default_pathext=None):
    """
    Returns the path extensions from environment or a default
    """
    if default_pathext is None:
        default_pathext = os.pathsep.join(['.COM', '.EXE', '.BAT', '.CMD'])
    return os.environ.get('PATHEXT', default_pathext)
